# pylint: disable=undefined-variable
from .jsonapi import *  # noqa: F403
from .utils import *  # noqa: F403
from .baseinfo import *  # noqa: F403
from .entries import *  # noqa: F403
from .index_metadb import *  # noqa: F403
from .links import *  # noqa: F403
from .optimade_json import *  # noqa: F403
from .references import *  # noqa: F403
from .responses import *  # noqa: F403
from .structures import *  # noqa: F403

__all__ = (
    jsonapi.__all__  # noqa: F405
    + utils.__all__  # noqa: F405
    + baseinfo.__all__  # noqa: F405
    + entries.__all__  # noqa: F405
    + index_metadb.__all__  # noqa: F405
    + links.__all__  # noqa: F405
    + optimade_json.__all__  # noqa: F405
    + references.__all__  # noqa: F405
    + responses.__all__  # noqa: F405
    + structures.__all__  # noqa: F405
)



# Copyright 2015-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa, Ask Hjorth Larsen, Jens Jørgen Mortensen
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
used with parser2.py, the new file format (aff)
"""
import numpy as np

parameters = {
    'mode': 'fd',
    'xc': 'LDA',
    'occupations': None,
    'poissonsolver': None,
    'h': None,  # Angstrom
    'gpts': None,
    'kpts': [(0.0, 0.0, 0.0)],
    'nbands': None,
    'charge': 0,
    'setups': {},
    'basis': {},
    'spinpol': None,
    'fixdensity': False,
    'filter': None,
    'mixer': None,
    'eigensolver': None,
    'background_charge': None,
    'external': None,
    'random': False,
    'hund': False,
    'maxiter': 333,
    'idiotproof': True,
    'symmetry': {'point_group': True,
                 'time_reversal': True,
                 'symmorphic': True,
                 'tolerance': 1e-7},
    'convergence': {'energy': 0.0005,  # eV / electron
                    'density': 1.0e-4,
                    'eigenstates': 4.0e-8,  # eV^2
                    'bands': 'occupied',
                    'forces': np.inf},  # eV / Ang
    'dtype': None,  # Deprecated
    'width': None,  # Deprecated
    'verbose': 0}





# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import KSpaceStyle


class none(KSpaceStyle):
    name = 'none'
    pass


class Ewald(KSpaceStyle):
    name = 'ewald'
    pass


class EwaldDisp(KSpaceStyle):
    name = 'ewald/disp'
    pass


class EwaldOmp(KSpaceStyle):
    name = 'ewald/omp'
    pass


class Pppm(KSpaceStyle):
    name = 'pppm'
    pass


class PppmCg(KSpaceStyle):
    name = 'pppm/cg'
    pass


class PppmDisp(KSpaceStyle):
    name = 'pppm/disp'
    pass


class PppmTip4p(KSpaceStyle):
    name = 'pppm/tip4p'
    pass


class PppmStagger(KSpaceStyle):
    name = 'pppm/stagger'
    pass


class PppmDispTip4p(KSpaceStyle):
    name = 'pppm/disp/tip4p'
    pass


class PppmGpu(KSpaceStyle):
    name = 'pppm/gpu'
    pass


class PppmOmp(KSpaceStyle):
    name = 'pppm/omp'
    pass


class PppmCgOmp(KSpaceStyle):
    name = 'pppm/cg/omp'
    pass


class PppmTip4pOmp(KSpaceStyle):
    name = 'pppm/tip4p/omp'
    pass


class Msm(KSpaceStyle):
    name = 'msm'
    pass


class MsmCg(KSpaceStyle):
    name = 'msm/cg'
    pass


class MsmOmp(KSpaceStyle):
    name = 'msm/omp'
    pass


class MsmCgOmp(KSpaceStyle):
    name = 'msm/cg/omp'
    pass




# Copyright 2016-2018 Micael Oliveira, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from builtins import object
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.simple_parser import mainFunction
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore import parser_backend
from nomadcore.unit_conversion import unit_conversion
from ase.data import chemical_symbols
from abinitparser.AbinitXC import ABINIT_NATIVE_IXC, ABINIT_LIBXC_IXC
import numpy as np
import re
import os
import logging
import time
import sys

try:
    basestring
except NameError:
    basestring = str

logger = logging.getLogger("nomad.ABINITParser")

parserInfo = {
  "name": "ABINIT_parser",
  "version": "1.0"
}

ABINIT_GEO_OPTIMIZATION = {
    1: "viscous_damped_md",
    2: "bfgs",
    3: "bfgs",
    4: "conjugate_gradient",
    5: "steepest_descent",
    7: "quenched_md",
    10: "dic_bfgs",
    11: "dic_bfgs",
    20: "diis"
}

class ABINITContext(object):
    """context for the sample parser"""

    def __init__(self):
        self.parser = None
        self.current_dataset = None
        self.abinitVars = None
        self.input = None
        self.initial_simulation_cell = None
        self.inputGIndex = None
        self.methodGIndex = None
        self.systemGIndex = None
        self.samplingGIndex = None
        self.basisGIndex = None
        self.frameSequence = []

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
        self.current_dataset = 0
        # Initialize dict to store Abinit variables. Two of them are created by default:
        #  - dataset "0", which will contain the values that are common to all datasets
        #  - dataset "1", as this is the default dataset number used by Abinit when the user
        #    does not specify the dataset number
        self.abinitVars = {key: {} for key in [0, 1]}
        self.input = None
        self.initial_simulation_cell = None
        self.inputGIndex = None
        self.methodGIndex = None
        self.systemGIndex = None
        self.samplingGIndex = None
        self.basisGIndex = None
        self.frameSequence = []

    def startedParsing(self, filename, parser):
        """called when parsing starts"""
        self.parser = parser
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def onClose_section_run(self, backend, gIndex, section):
        """Trigger called when section_run is closed.
        """
        if section["x_abinit_completed"] is not None:
            backend.addValue("run_clean_end", True)
        # Convert date and time to epoch time
        if (section["x_abinit_start_date"] is not None) and (section["x_abinit_start_time"] is not None):
            abi_time = time.strptime(str("%s %s") % (section["x_abinit_start_date"][-1],
                                                     section["x_abinit_start_time"][-1]), "%a %d %b %Y %Hh%M")
            backend.addValue("time_run_date_start", time.mktime(abi_time))

    def onOpen_x_abinit_section_dataset(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_dataset is opened.
        """
        self.samplingGIndex = backend.openSection("section_sampling_method")

    def onClose_x_abinit_section_dataset(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_dataset is closed.
        """
        if len(self.frameSequence) > 1:
            frameGIndex = backend.openSection("section_frame_sequence")
            if section["x_abinit_geometry_optimization_converged"] is not None:
                if section["x_abinit_geometry_optimization_converged"][-1] == "are converged":
                    backend.addValue("geometry_optimization_converged", True)
                else:
                    backend.addValue("geometry_optimization_converged", False)
            backend.closeSection("section_frame_sequence", frameGIndex)
        self.frameSequence = []
        backend.closeSection("section_sampling_method", self.samplingGIndex)

    def onClose_section_frame_sequence(self, backend, gIndex, section):
        """Trigger called when section_framce_sequence is closed.
        """
        backend.addValue("number_of_frames_in_sequence", len(self.frameSequence))
        backend.addArrayValues("frame_sequence_local_frames_ref", np.array(self.frameSequence))
        backend.addValue("frame_sequence_to_sampling_ref", self.samplingGIndex)

    def onClose_section_sampling_method(self, backend, gIndex, section):
        """Trigger called when section_sampling_method is closed.
        """
        if self.input["x_abinit_var_ionmov"] is not None:
            ionmov = self.input["x_abinit_var_ionmov"][-1]
            if ionmov in [2, 3, 4, 5, 7, 10, 11, 20] or (ionmov == 1 and self.input["x_abinit_var_vis"] > 0.0):
                sampling_method = "geometry_optimization"
                backend.addValue("geometry_optimization_method", ABINIT_GEO_OPTIMIZATION[ionmov])
            elif ionmov in [6, 8, 12, 13, 14, 23] or (ionmov == 1 and self.input["x_abinit_var_vis"] == 0.0):
                sampling_method = "molecular_dynamics"
            elif ionmov == 9:
                sampling_method = "langevin_dynamics"
            else:
                sampling_method = ""
            backend.addValue("sampling_method", sampling_method)

        if self.input["x_abinit_var_tolmxf"] is not None:
            backend.addValue("geometry_optimization_threshold_force",
                             unit_conversion.convert_unit(self.input["x_abinit_var_tolmxf"][-1], "hartree / bohr"))
        if self.input["x_abinit_var_tolmxde"] is not None:
            backend.addValue("geometry_optimization_energy_change",
                             unit_conversion.convert_unit(self.input["x_abinit_var_tolmxde"][-1], "hartree"))

    def onOpen_section_single_configuration_calculation(self, backend, gIndex, section):
        """Trigger called when section_single_configuration_calculation is opened.
        """
        self.systemGIndex = backend.openSection("section_system")
        backend.addValue("single_configuration_calculation_to_system_ref", self.systemGIndex)
        backend.addValue("single_configuration_to_calculation_method_ref", self.methodGIndex)
        self.frameSequence.append(gIndex)

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        """Trigger called when section_single_configuration_calculation is closed.
        """
        backend.closeSection("section_system", self.systemGIndex)

        converged = False
        if section["x_abinit_single_configuration_calculation_converged"] is not None:
            if section["x_abinit_single_configuration_calculation_converged"][-1] == "is converged" or \
               section["x_abinit_single_configuration_calculation_converged"][-1] == "converged" or \
               section["x_abinit_single_configuration_calculation_converged"][-1] == "are converged":
                converged = True
        backend.addValue("single_configuration_calculation_converged", converged)

        if section["x_abinit_energy_xc"] is not None:
            backend.addValue("energy_XC", unit_conversion.convert_unit(section["x_abinit_energy_xc"][-1], "hartree"))
        if section["x_abinit_energy_kinetic"] is not None:
            backend.addValue("electronic_kinetic_energy",
                             unit_conversion.convert_unit(section["x_abinit_energy_kinetic"][-1], "hartree"))

        if section["energy_total"] is None and section["x_abinit_energy_total"] is not None:
            backend.addValue("energy_total",
                             unit_conversion.convert_unit(section["x_abinit_energy_total"][-1], "hartree"))

        if section["x_abinit_atom_force"] is not None:
            atom_forces_list = section["x_abinit_atom_force"]
        elif section["x_abinit_atom_force_final"] is not None:
            atom_forces_list = section["x_abinit_atom_force_final"]
        else:
            atom_forces_list = None

        if atom_forces_list is not None:
            atom_forces = backend.arrayForMetaInfo("atom_forces_raw", [self.input["x_abinit_var_natom"][-1],3])
            n_atom = 0
            for force_string in atom_forces_list:
                for dir in range(3):
                    atom_forces[n_atom, dir] = unit_conversion.convert_unit(float(force_string.split()[dir]), "hartree / bohr")
                n_atom += 1
            backend.addArrayValues("atom_forces_raw", atom_forces)

        if section["x_abinit_fermi_energy"] is not None:
            backend.addArrayValues("energy_reference_fermi",
                                   np.array([unit_conversion.convert_unit(section["x_abinit_fermi_energy"][-1], "hartree")]))

    def onClose_section_eigenvalues(self, backend, gIndex, section):
        """Trigger called when section_eigenvalues is closed.
        """
        nkpt = int(self.input["x_abinit_var_nkpt"][-1])
        nspin = int(self.input["x_abinit_var_nsppol"][-1])
        nband = int(self.input["x_abinit_var_nband"][-1][0][0])
        for ispin in range(nspin):
            for ikpt in range(nkpt):
                if self.input["x_abinit_var_nband"][-1][ikpt][ispin] != nband:
                    nband = 0
                    break

        if nband == 0:
            logger.warn("Number of bands in this calculation is k-point dependent.")
        else:
            if section["x_abinit_eigenvalues"] is not None:
                if len(section["x_abinit_eigenvalues"]) == 1:
                    ev_string = section["x_abinit_eigenvalues"][0]
                else:
                    ev_string = " ".join(section["x_abinit_eigenvalues"])
                abi_eigenvalues = ev_string.split()
            else:
                abi_eigenvalues = []
                logger.warn("Eigenvalues are not available.")
            eigenvalues = np.array([unit_conversion.convert_unit(float(x), "hartree") for x in abi_eigenvalues])

            if section["x_abinit_occupations"] is not None:
                if len(section["x_abinit_occupations"]) == 1:
                    occs_string = section["x_abinit_occupations"][0]
                else:
                    occs_string = " ".join(section["x_abinit_occupations"])
                abi_occs = occs_string.split()
            elif self.input["x_abinit_var_occ"][-1] is not None:
                abi_occs = self.input["x_abinit_var_occ"][-1]
            else:
                abi_occs = []
                logger.warn("Occupations are not available.")
            occupations = np.array([float(x) for x in abi_occs])

            abi_kpoints = []
            if section["x_abinit_kpt"] is not None:
                for kpt in section["x_abinit_kpt"]:
                    abi_kpoints.append([float(x) for x in kpt.split()])
            elif self.input["x_abinit_var_kpt"] is not None:
                for ikpt in range(nkpt):
                    abi_kpoints.append([float(x) for x in self.input["x_abinit_var_kpt"][-1][ikpt]])
            else:
                logger.warn("K-points are not available.")
            kpoints = np.array(abi_kpoints)

            if len(kpoints) == nkpt and len(eigenvalues) == nband*nspin*nkpt and len(occupations) == nband * nspin * nkpt:
                backend.addValue("eigenvalues_kind", "normal")
                backend.addValue("number_of_eigenvalues", nband)
                backend.addValue("number_of_eigenvalues_kpoints", nkpt)
                backend.addArrayValues("eigenvalues_kpoints", kpoints)
                backend.addArrayValues("eigenvalues_values", eigenvalues.reshape([nspin, nkpt, nband]))
                backend.addArrayValues("eigenvalues_occupation", occupations.reshape([nspin, nkpt, nband]))

    def onOpen_x_abinit_section_dataset_header(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_dataset is opened.
        """
        self.inputGIndex = backend.openSection("x_abinit_section_input")

    def onClose_x_abinit_section_dataset_header(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_dataset is closed.
        """
        self.current_dataset = section["x_abinit_dataset_number"][-1]
        backend.closeSection("x_abinit_section_input", self.inputGIndex)

        self.basisGIndex = backend.openSection("section_basis_set_cell_dependent")
        backend.addValue("basis_set_cell_dependent_kind", "plane_waves")
        if self.input["x_abinit_var_ecut"] is not None:
            backend.addValue("basis_set_planewave_cutoff",
                             unit_conversion.convert_unit(self.input["x_abinit_var_ecut"][-1], 'hartree'))
        backend.addValue("basis_set_cell_dependent_name",
                         "PW_%s"%(unit_conversion.convert_unit(self.input["x_abinit_var_ecut"][-1], 'hartree', 'rydberg')))
        backend.closeSection("section_basis_set_cell_dependent", self.basisGIndex)

        self.methodGIndex = backend.openSection("section_method")
        backend.closeSection("section_method", self.methodGIndex)

        self.initial_simulation_cell = []
        for axis in [1, 2, 3]:
            # FAIRD edit (dts@physik.hu-berlin.de), the split is failing for negative
            # decimals in the third column. It gives "2.7 2.7-2.7" for some reason.
            # We need to add a whitespace to split nicely.
            self.initial_simulation_cell.append(
               section["x_abinit_vprim_%s" % axis][-1].replace('-',' -').split())

    def onClose_section_method(self, backend, gIndex, section):
        """Trigger called when section_method is closed.
        """
        backend.addValue("stress_tensor_method", "analytic")
        backend.addValue("number_of_spin_channels", self.input["x_abinit_var_nsppol"][-1])
        backend.addValue("scf_max_iteration", self.input["x_abinit_var_nstep"][-1])
        if self.input["x_abinit_var_toldfe"] is not None:
            backend.addValue("scf_threshold_energy_change",
                             unit_conversion.convert_unit(self.input["x_abinit_var_toldfe"][-1], 'hartree'))
        backend.addValue("self_interaction_correction_method", "")
        if self.input["x_abinit_var_occopt"][-1] == 3:
            smear_kind = "fermi"
        elif self.input["x_abinit_var_occopt"][-1] == 4 or self.input["x_abinit_var_occopt"] == 5:
            smear_kind = "marzari-vanderbilt"
        elif self.input["x_abinit_var_occopt"][-1] == 6:
            smear_kind = "methfessel-paxton"
        elif self.input["x_abinit_var_occopt"][-1] == 7:
            smear_kind = "gaussian"
        elif self.input["x_abinit_var_occopt"][-1] == 8:
            logger.error("Illegal value for Abinit input variable occopt")
            smear_kind = ""
        else:
            smear_kind = ""
        backend.addValue("smearing_kind", smear_kind)
        if self.input["x_abinit_var_tsmear"] is not None:
            backend.addValue("smearing_width",
                             unit_conversion.convert_unit(self.input["x_abinit_var_tsmear"][-1], 'hartree'))

        gIndex = backend.openSection("section_method_basis_set")
        backend.addValue("method_basis_set_kind", "wavefunction")
        backend.addValue("mapping_section_method_basis_set_cell_associated", self.basisGIndex)
        backend.closeSection("section_method_basis_set", gIndex)

        backend.addValue("electronic_structure_method", "DFT")
        ixc = int(self.input["x_abinit_var_ixc"][-1])
        if ixc >= 0:
            xc_functionals = ABINIT_NATIVE_IXC[str(ixc)]
        else:
            xc_functionals = []
            functional1 = -ixc//1000
            if functional1 > 0:
                xc_functionals.append(ABINIT_LIBXC_IXC[str(functional1)])
            functional2 = -ixc - (-ixc//1000)*1000
            if functional2 > 0:
                xc_functionals.append(ABINIT_LIBXC_IXC[str(functional2)])

        if xc_functionals is not None:
            for xc_functional in xc_functionals:
                gIndex = backend.openSection('section_XC_functionals')
                for key, value in sorted(xc_functional.items()):
                    if isinstance(value, (list, dict)):
                        backend.addValue(key, value)
                    else:
                        backend.addValue(key, value)
                backend.closeSection('section_XC_functionals', gIndex)

    def onClose_section_system(self, backend, gIndex, section):
        """Trigger called when section_system is closed.
        """
        species_count = {}
        for z in self.input["x_abinit_var_znucl"][-1]:
            species_count[chemical_symbols[int(z)]] = 0

        atom_types = []
        for z in self.input["x_abinit_var_znucl"][-1]:
            symbol = chemical_symbols[int(z)]
            species_count[symbol] += 1
            if species_count[symbol] > 1:
               atom_type = symbol + str(species_count[symbol])
            else:
               atom_type = symbol
            atom_types.append(atom_type)

        # Grabs an array with the dtype of meta name atom_labels and
        # shape of self.input('x_abinit..').
        atom_labels = backend.arrayForMetaInfo("atom_labels", self.input["x_abinit_var_natom"][-1])

        for atom_index in range(self.input["x_abinit_var_natom"][-1]):
            atom_labels[atom_index] = atom_types[self.input["x_abinit_var_typat"][-1][atom_index] - 1]

        backend.addArrayValues("atom_labels", atom_labels)

        if section["x_abinit_atom_xcart"] is not None:
            atom_xcart_list = section["x_abinit_atom_xcart"]
            xcart_unit = "bohr"
        elif section["x_abinit_atom_xcart_final"] is not None:
            atom_xcart_list = section["x_abinit_atom_xcart_final"]
            xcart_unit = "angstrom"
        else:
            atom_xcart_list = None
        n_atom = self.input["x_abinit_var_natom"][-1]
        if atom_xcart_list is not None:
            atom_xcart = backend.arrayForMetaInfo("atom_positions", [n_atom, 3])
            for iatom in range(n_atom):
                for dir in range(3):
                    atom_xcart[iatom, dir] = float(atom_xcart_list[iatom].split()[dir])
        else:
            xcart_unit = "bohr"
            if self.input["x_abinit_var_xcart"] is not None:
                atom_xcart = self.input["x_abinit_var_xcart"][-1]
            elif n_atom == 1:
                atom_xcart = np.array([[0, 0, 0]])
            else:
                logger.error("Positions of atoms is not available")
        for iatom in range(n_atom):
            for dir in range(3):
                atom_xcart[iatom, dir] = unit_conversion.convert_unit(atom_xcart[iatom, dir], xcart_unit)
        backend.addArrayValues("atom_positions", atom_xcart)


        backend.addArrayValues("configuration_periodic_dimensions", np.array([True, True, True]))

        backend.addValue("number_of_atoms", self.input["x_abinit_var_natom"][-1])

        #backend.addValue("space_group_3D_number", self.input["x_abinit_var_spgroup"][-1]) # leave this to the normalizer...

        vprim = backend.arrayForMetaInfo("simulation_cell", [3, 3])
        for axis in [1, 2, 3]:
            if section["x_abinit_vprim_%s"%axis] is None:
                vprim[axis-1] = self.initial_simulation_cell[axis-1]
            else:
                vprim[axis-1] = section["x_abinit_vprim_%s"%axis][-1].split()
            for component in range(3):
                vprim[axis-1][component] = unit_conversion.convert_unit(vprim[axis-1][component], 'bohr')
        backend.addArrayValues("simulation_cell", vprim)

    def onOpen_x_abinit_section_input(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_input is opened.
        """
        self.input = section

    def onClose_x_abinit_section_input(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_input is closed.
        """
        dataset_vars = {}
        for varname in backend.superBackend.metaInfoEnv().infoKinds.keys():
            if "x_abinit_var_" in varname:
                dataset_vars[varname] = None

        dataset_vars.update(self.abinitVars[0])
        dataset_vars.update(self.abinitVars[self.current_dataset])

        # Take care of default values. We need to do this here because the default values of some variables depend on
        # the value of other variables.
        if dataset_vars["x_abinit_var_ntypat"] is None:
            dataset_vars["x_abinit_var_ntypat"] = 1
        if dataset_vars["x_abinit_var_npsp"] is None:
            dataset_vars["x_abinit_var_npsp"] = dataset_vars["x_abinit_var_ntypat"]
        if dataset_vars["x_abinit_var_nshiftk"] is None:
            dataset_vars["x_abinit_var_nshiftk"] = 1
        if dataset_vars["x_abinit_var_natrd"] is None:
            dataset_vars["x_abinit_var_natrd"] = dataset_vars["x_abinit_var_natom"]
        if dataset_vars["x_abinit_var_nsppol"] is None:
            dataset_vars["x_abinit_var_nsppol"] = 1
        if dataset_vars["x_abinit_var_nspden"] is None:
            dataset_vars["x_abinit_var_nspden"] = dataset_vars["x_abinit_var_nsppol"]
        if dataset_vars["x_abinit_var_nkpt"] is None:
            dataset_vars["x_abinit_var_nkpt"] = 1
        if dataset_vars["x_abinit_var_occopt"] is None:
            dataset_vars["x_abinit_var_occopt"] = 1
        if dataset_vars["x_abinit_var_ixc"] is None:
            dataset_vars["x_abinit_var_ixc"] = 1

        # Fix nband
        if len(dataset_vars["x_abinit_var_nband"].split()) == 1:
            nband = ""
            for ispin in range(int(dataset_vars["x_abinit_var_nsppol"])):
                for ikpt in range(int(dataset_vars["x_abinit_var_nkpt"])):
                    nband += dataset_vars["x_abinit_var_nband"]+" "
            dataset_vars["x_abinit_var_nband"] = nband

        for varname, varvalue in dataset_vars.items():

            meta_info = backend.superBackend.metaInfoEnv().infoKindEl(varname)

            # Skip optional variables that do not have a value or that are not defined in the meta-info
            if varvalue is None or meta_info is None:
                continue

            if varname == "x_abinit_var_occ":
                # Abinit allows for different numbers of bands per k-point and/or spin channel
                # This means the occupations need to be handled in a special way

                if dataset_vars["x_abinit_var_occopt"] != 2 and dataset_vars["x_abinit_var_nsppol"] == 1:
                    # In this case Abinit only prints the occupations for one k-point, as occupations are the same
                    # for all k-points
                    varvalue = ""
                    for ikpt in range(int(dataset_vars["x_abinit_var_nkpt"])):
                        varvalue += dataset_vars["x_abinit_var_occ"]+" "

                nband = sum([int(x) for x in dataset_vars["x_abinit_var_nband"].split()])
                array = np.array(varvalue.split(), dtype=parser_backend.numpyDtypeForDtypeStr(meta_info.dtypeStr))
                backend.addArrayValues(varname, array.reshape([nband]))
            elif varname == "x_abinit_var_ixc":
                # If no value of ixc is given in the input file, Abinit will try to choose it from the pseudopotentials.
                # Since the pseudopotentials are read while performing the calculations for a given dataset, ixc might
                # have been already read and stored. In that case we ignore the value stored in dataset_vars.
                if section["x_abinit_var_ixc"] is None:
                    backend.addValue(varname, backend.convertScalarStringValue(varname, varvalue))

            elif len(meta_info.shape) == 0:
                # This is a simple scalar
                backend.addValue(varname, backend.convertScalarStringValue(varname, varvalue))

            else:
                # This is an array
                array = np.array(varvalue.split(), dtype=parser_backend.numpyDtypeForDtypeStr(meta_info.dtypeStr))
                shape = []
                for dim in meta_info.shape:
                    if isinstance(dim, basestring):
                        # Replace all instances of Abinit variables that appear in the dimension
                        # with their actual values.
                        dim_regex = '(?P<abi_var>x_abinit_var_\w+)'
                        for mo in re.finditer(dim_regex, dim):
                            dim = re.sub(mo.group("abi_var"), str(dataset_vars[mo.group("abi_var")]), dim)
                        # In some cases the dimension is given as a numerical expression that needs to be evaluated
                        try:
                           dim = eval(dim)
                        except SyntaxError as e:
                           print(dim, meta_info.name, meta_info.shape)
                           raise e
                    shape.append(dim)
                backend.addArrayValues(varname, array.reshape(shape))

    def onClose_x_abinit_section_stress_tensor(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_stress_tensor is closed.
        """
        xx = unit_conversion.convert_unit(section["x_abinit_stress_tensor_xx"][-1], "hartree/bohr^3")
        yy = unit_conversion.convert_unit(section["x_abinit_stress_tensor_yy"][-1], "hartree/bohr^3")
        zz = unit_conversion.convert_unit(section["x_abinit_stress_tensor_zz"][-1], "hartree/bohr^3")
        zy = unit_conversion.convert_unit(section["x_abinit_stress_tensor_zy"][-1], "hartree/bohr^3")
        zx = unit_conversion.convert_unit(section["x_abinit_stress_tensor_zx"][-1], "hartree/bohr^3")
        yx = unit_conversion.convert_unit(section["x_abinit_stress_tensor_yx"][-1], "hartree/bohr^3")
        backend.addArrayValues("stress_tensor", np.array([[xx, yx, zx], [yx, yy, zy], [zx, zy, zz]]))

    def onClose_x_abinit_section_var(self, backend, gIndex, section):
        """Trigger called when x_abinit_section_var is closed.
        """
        # We store all the variables read in a dictionary for latter use.
        m = re.search('[0-9]+$', section["x_abinit_varname"][0])
        if m is None:
            dataset = 0
        else:
            dataset = int(m.group(0))
            section["x_abinit_varname"][0] = re.sub('[0-9]+$', '', section["x_abinit_varname"][0])

        if dataset not in self.abinitVars.keys():
            self.abinitVars[dataset] = {}
        if len(section["x_abinit_varvalue"]) == 1:
            self.abinitVars[dataset]["x_abinit_var_" + section["x_abinit_varname"][0]] = section["x_abinit_varvalue"][0]
        else:
            # We have an array of values. We will only set the variable if we have all the values, that is, if the array
            # was not truncated.
            if section["x_abinit_vartruncation"] is None:
                self.abinitVars[dataset]["x_abinit_var_" + section["x_abinit_varname"][0]] = \
                    " ".join(section["x_abinit_varvalue"])


# description of the input
headerMatcher = \
    SM(name='Header',
       startReStr=r"\.Version\s*[0-9a-zA-Z_.]*\s*of ABINIT\s*$",
       required=True,
       forwardMatch=True,
       subMatchers=[SM(r"\.Version (?P<program_version>[0-9a-zA-Z_.]*) of ABINIT\s*$"),
                    SM(r"\.\((?P<x_abinit_parallel_compilation>[a-zA-Z]*)\s*version, prepared for a "
                       r"(?P<program_compilation_host>\S*)\s*computer\)\s*$"),
                    SM(r"\s*$"),
                    SM(startReStr="\.Copyright \(C\) 1998-[0-9]{4} ABINIT group .\s*$",
                       coverageIgnore=True,
                       subMatchers=[SM(r"\s*ABINIT comes with ABSOLUTELY NO WARRANTY.\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*It is free software, and you are welcome to redistribute it\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*under certain conditions \(GNU General Public License,\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*see ~abinit/COPYING or http://www.gnu.org/copyleft/gpl.txt\).\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*ABINIT is a project of the Universite Catholique de Louvain,\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Corning Inc. and other collaborators, see "
                                       r"~abinit/doc/developers/contributors.txt .\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Please read ~abinit/doc/users/acknowledgments.html for suggested\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*acknowledgments of the ABINIT effort.\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*For more information, see http://www.abinit.org .\s*$",
                                       coverageIgnore=True)
                                    ]
                       ),
                    SM(r"\.Starting date : (?P<x_abinit_start_date>[0-9a-zA-Z ]*)\.\s*$"),
                    SM(r"^- \( at\s*(?P<x_abinit_start_time>[0-9a-z]*)\s*\)\s*$"),
                    SM(r"^- input  file\s*->\s*(?P<x_abinit_input_file>\S*)\s*$"),
                    SM(r"^- output file\s*->\s*(?P<x_abinit_output_file>\S*)\s*$"),
                    SM(r"^- root for input  files\s*->\s*(?P<x_abinit_input_files_root>\S*)\s*$"),
                    SM(r"^- root for output files\s*->\s*(?P<x_abinit_output_files_root>\S*)\s*$")
                    ],
       )

timerMatcher = \
    SM(name='Timer',
       startReStr="- Total cpu\s*time",
       endReStr="={80}",
       required=True,
       forwardMatch=True,
       coverageIgnore=True,
       subMatchers=[SM(r"- Total cpu\s*time\s*\(\S*\):\s*(?P<x_abinit_total_cpu_time>[0-9.]+)\s*\S*\s*\S*\s*$"),
                    SM(r"- Total wall clock time\s*\(\S*\):\s*(?P<x_abinit_total_wallclock_time>[0-9.]+)\s*\S*\s*\S*\s*$"),
                    SM(r"-\s*$",
                       coverageIgnore=True),
                    SM(name="Profiling",
                       startReStr="- For major independent code sections, cpu and wall times \(sec\),",
                       endReStr="- subtotal(\s*[0-9.]+){4}",
                       repeats=True,
                       coverageIgnore=True,
                       subMatchers=[SM(r"-\s*as well as % of the (total time and number of calls|time and number of "
                                       r"calls for node [0-9]+-)",
                                       coverageIgnore=True),
                                    SM(r"-<BEGIN_TIMER mpi_nprocs = [0-9]+, omp_nthreads = [0-9]+, mpi_rank = "
                                       r"([0-9]+|world)>",
                                       coverageIgnore=True),
                                    SM(r"- cpu_time =\s*[0-9.]+, wall_time =\s*[0-9.]+",
                                       coverageIgnore=True),
                                    SM(r"-",
                                       coverageIgnore=True),
                                    SM(r"- routine\s*cpu\s*%\s*wall\s*%\s*number of calls\s*Gflops",
                                       coverageIgnore=True),
                                    SM(r"-\s*\(-1=no count\)",
                                       coverageIgnore=True),
                                    SM(r"-(\s*\S*)+(\s*[-0-9.]+){6}",
                                       coverageIgnore=True, repeats=True),
                                    SM(r"-<END_TIMER>",
                                       coverageIgnore=True),
                                    SM(r"-",
                                       coverageIgnore=True, required=False),
                                    ]
                       )
                    ]
       )

memestimationMatcher = \
    SM(name='MemEstimation',
       startReStr=r"\s*(Symmetries|DATASET\s*[0-9]{1,4})\s*:\s*space group\s*(\S*\s*){1,3}\s*\(\#\s*[0-9]{1,3}\);\s*"
                  r"Bravais\s*\S*\s*\([0-9a-zA-Z- .]*\)\s*$",
       endReStr=r"={80}",
       repeats=True,
       subMatchers=[SM(r"={80}",
                       coverageIgnore=True),
                    SM(r"\s*Values of the parameters that define the memory need (of the present run|for DATASET\s*"
                       r"[0-9]+\.)",
                       coverageIgnore=True),
                    # We ignore the variables printed here, as what is printed is Abinit version dependent and depends
                    # on the actual values of multiple parameters. The most important variables are repeated later.
                    SM(r"(-|P)?(\s*\S*\s*=\s*[0-9]+)+",
                       coverageIgnore=True, repeats=True),
                    SM(r"={80}",
                       coverageIgnore=True),
                    SM(r"P This job should need less than\s*[0-9.]+\s*Mbytes of memory.",
                       coverageIgnore=True),
                    SM(r"\s*Rough estimation \(10\% accuracy\) of disk space for files :",
                       coverageIgnore=True),
                    SM(r"_ WF disk file :\s*[0-9.]+\s*Mbytes ; DEN or POT disk file :\s*[0-9.]+\s*Mbytes.",
                       coverageIgnore=True),
                    SM(r"={80}",
                       coverageIgnore=True)
                    ]
       )

inputVarsMatcher = \
    SM(name='InputVars',
       startReStr=r"-{80}",
       endReStr=r"={80}",
       required=True,
       coverageIgnore=True,
       subMatchers=[SM(r"-{13} Echo of variables that govern the present computation -{12}",
                       coverageIgnore=True),
                    SM(r"-{80}",
                       coverageIgnore=True),
                    SM(r"-",
                       coverageIgnore=True),
                    SM(r"- outvars: echo of selected default values",
                       coverageIgnore=True),
                    SM(r"-(\s*\w+\s*=\s*[0-9]+\s*,{0,1})*"),
                    SM(r"-",
                       coverageIgnore=True),
                    SM(r"- outvars: echo of global parameters not present in the input file",
                       coverageIgnore=True),
                    SM(r"-(\s*\w+\s*=\s*[0-9]+\s*,{0,1})*"),
                    SM(r"-",
                       coverageIgnore=True),
                    SM(r" -outvars: echo values of preprocessed input variables --------",
                       coverageIgnore=True),
                    SM(startReStr=r"[-P]?\s+[_a-zA-Z0-9]+\s+(\S*\s*)+\s*(Hartree|Bohr)?\s*$",
                       forwardMatch=True,
                       sections=['x_abinit_section_var'],
                       repeats=True,
                       subMatchers=[SM(r"[-P]?\s+(?P<x_abinit_varname>[_a-zA-Z0-9]+)\s+"
                                       r"(?P<x_abinit_varvalue>((?!(Bohr|Hartree))\S*\s*)+)\s*(Hartree|Bohr)?\s*$"),
                                    SM(r"\s{20,}(?P<x_abinit_varvalue>([0-9Ee.+-]+\s*)+)\s*$",
                                       repeats=True),
                                    SM(r"\s{20,}outvar(_i_n|s)\s*: Printing only first\s*"
                                       r"(?P<x_abinit_vartruncation>[0-9]+)\s*[-a-zA-Z]*.\s*$",
                                       required=False)
                                    ]
                       ),
                    SM(r"={80}",
                       coverageIgnore=True),
                    SM(r"\s*chkinp: Checking input parameters for consistency(\.|,\s*jdtset=\s*[0-9]+\.)",
                       coverageIgnore=True, repeats=True)
                    ]
       )


eigenvaluesBlockMatcher = \
    SM(name="EigenvaluesBlock",
       startReStr=r"\s*Eigenvalues \(hartree\) for nkpt=\s*[0-9]+\s*k points(, SPIN (UP|DOWN))?:\s*$",
       repeats=True,
       subMatchers=[SM(startReStr=r"\s*kpt#\s*[0-9]+, nband=\s*[0-9]+, wtk=\s*[0-9.]+\s*, kpt=(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){3}"
                                  r"\s*\(reduced coord\)\s*$",
                       forwardMatch=True,
                       repeats=True,
                       subMatchers=[SM(r"\s*kpt#\s*[0-9]+, nband=\s*[0-9]+, wtk=\s*(?P<x_abinit_wtk>[0-9.]+)\s*, "
                                       r"kpt=(?P<x_abinit_kpt>(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){3})\s*\(reduced coord\)\s*$"),
                                    SM(r"(?P<x_abinit_eigenvalues>(\s+[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)+)\s*$",
                                       repeats=True),
                                    SM(r"\s*occupation numbers for kpt#\s+[0-9]+\s*$",
                                       required=False),
                                    SM(r"(?P<x_abinit_occupations>(\s+[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)+)\s*$",
                                       repeats=True, required=False)
                                    ]
                       ),
                    ]
       )

SCFResultsMatcher = \
    SM(name='SCFResults',
       startReStr=r"\s*----iterations are completed or convergence reached----\s*$",
       required=False,
       subMatchers=[SM(r"\s*Mean square residual over all n,k,spin=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*;\s*max=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True),
                    SM(startReStr=r"\s*([-+0-9.]+\s*){3}\s*[0-9]?\s*[-+0-9eEdD.]+\s*kpt; spin; max resid\(k\); each band:\s*$",
                       coverageIgnore=True,
                       repeats=True,
                       subMatchers=[SM(r"\s*([-+0-9eEdD.]+\s*)+\s*$",
                                       coverageIgnore=True)
                                    ]
                       ),
                    SM(startReStr=r"\s*reduced coordinates \(array xred\) for\s*[0-9]+\s*atoms\s*$",
                       coverageIgnore=True,
                       subMatchers=[SM(r"\s*([-+0-9.]+\s*){3}\s*$",
                                       coverageIgnore=True, repeats=True)
                                    ]
                       ),
                    SM(startReStr=r"\s*rms dE/dt=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*; max dE/dt=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*; dE/dt below "
                                  r"\(all hartree\)\s*$",
                       coverageIgnore=True,
                       subMatchers=[SM(r"\s*[0-9]+\s*([-+0-9.]+\s*){3}\s*$",
                                       coverageIgnore=True, repeats=True)
                                    ]
                       ),
                    SM(startReStr=r"\s*cartesian coordinates \(angstrom\) at end:\s*$",
                       subMatchers=[SM(r"\s*[0-9]+(?P<x_abinit_atom_xcart_final>(\s*[-+0-9.]+){3})\s*$",
                                       repeats=True)
                                    ]
                       ),
                    SM(startReStr=r"\s*cartesian forces \(hartree/bohr\) at end:\s*$",
                       subMatchers=[SM(r"\s*[0-9]+(?P<x_abinit_atom_force_final>(\s*[-+0-9.]+){3})\s*$",
                                       repeats=True),
                                    SM(r"\s*frms,max,avg=(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){5}\s*h/b\s*$")
                                    ]
                       ),
                    SM(startReStr=r"\s*cartesian forces \(eV/Angstrom\) at end:\s*$",
                       subMatchers=[SM(r"\s*[0-9]+(\s*[-+0-9.]+){3}\s*$",
                                       repeats=True),
                                    SM(r"\s*frms,max,avg=(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){5}\s*e/A\s*$")
                                    ]
                       ),
                    SM(r"\s*length scales=(\s*[0-9.]+){3}\s*bohr\s*$",
                       coverageIgnore=True),
                    SM(r"\s*=(\s*[0-9.]+){3}\s*angstroms\s*$",
                       coverageIgnore=True),
                    SM(r"\s*prteigrs : about to open file\s*(?P<x_abinit_eig_filename>\S*)\s*$",
                       required=False),
                    SM(r"\s*Fermi \(or HOMO\) energy \(hartree\) =\s*(?P<x_abinit_fermi_energy>[-+0-9.]+)\s*"
                       r"Average Vxc \(hartree\)=\s*[-+0-9.]+\s*$"),
                    SM(r"\s*Magnetisation \(Bohr magneton\)=\s*(?P<x_abinit_magnetisation>[-+0-9.eEdD]*)\s*$"),
                    SM(r"\s*Total spin up =\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*Total spin down =\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True),
                    SM(name="Eigenvalues",
                       startReStr=r"\s*Eigenvalues \(hartree\) for nkpt=\s*[0-9]+\s*k points(, SPIN (UP|DOWN))?:\s*$",
                       forwardMatch=True,
                       sections=["section_eigenvalues"],
                       subMatchers=[eigenvaluesBlockMatcher]
                       ),
                    SM(startReStr=r"\s*(Total charge density|Spin up density|Spin down density|"
                                  r"Magnetization \(spin up - spin down\)|"
                                  r"Relative magnetization \(=zeta, between -1 and 1\))\s*(\[el/Bohr\^3\])?\s*$",
                       coverageIgnore=True,
                       repeats=True,
                       subMatchers=[SM(r",(Next)?\s*(m|M)(axi|ini)mum=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*at reduced coord."
                                       r"(\s*[0-9.]+){3}\s*$",
                                       repeats=True),
                                    SM(r",\s*Integrated=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$"),
                                    ]
                       ),
                    SM(r"-{80}",
                       coverageIgnore=True),
                    SM(r"\s*Components of total free energy \(in Hartree\) :\s*$",
                       coverageIgnore=True),
                    SM(r"\s*Kinetic energy\s*=\s*(?P<x_abinit_energy_kinetic>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*Hartree energy\s*=\s*(?P<x_abinit_energy_hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*XC energy\s*=\s*(?P<x_abinit_energy_xc>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*Ewald energy\s*=\s*(?P<x_abinit_energy_ewald>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*PspCore energy\s*=\s*(?P<x_abinit_energy_psp_core>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*Loc. psp. energy\s*=\s*(?P<x_abinit_energy_psp_local>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*NL   psp  energy\s*=\s*(?P<x_abinit_energy_psp_nonlocal>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*>{5}\s*Internal E=\s*(?P<x_abinit_energy_internal>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*-kT\*entropy\s*=\s*(?P<x_abinit_energy_ktentropy>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*>{9}\s*Etotal=\s*(?P<x_abinit_energy_total>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*Other information on the energy :\s*$",
                       coverageIgnore=True),
                    SM(r"\s*Total energy\(eV\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*;\s*Band energy \(Ha\)=\s*"
                       r"(?P<x_abinit_energy_band>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"-{80}",
                       coverageIgnore=True),
                    SM(startReStr=r"\s*rms coord change=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*atom, delta coord \(reduced\):\s*$",
                       subMatchers=[SM(r"\s*\d+(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){3}\s*$",
                                       repeats=True),
                                    ]),
                    SM(r"\s*Cartesian components of stress tensor \(hartree/bohr\^3\)\s*$",
                       coverageIgnore=True),
                    SM(r"\s*sigma\(1 1\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*sigma\(3 2\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True),
                    SM(r"\s*sigma\(2 2\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*sigma\(3 1\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True),
                    SM(r"\s*sigma\(3 3\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*sigma\(2 1\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True),
                    SM(r"-Cartesian components of stress tensor \(GPa\)\s*\[Pressure=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*GPa]\s*$",
                       coverageIgnore=True),
                    SM(r"- sigma\(1 1\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*sigma\(3 2\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True),
                    SM(r"- sigma\(2 2\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*sigma\(3 1\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True),
                    SM(r"- sigma\(3 3\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*sigma\(2 1\)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       coverageIgnore=True)
                    ]
       )

SCFOutput = \
    SM(name='SCFOutput',
       startReStr=r"-{3}OUTPUT-{71}\s*$",
       required=False,
       subMatchers=[SM(startReStr=r"\s*Cartesian coordinates \(xcart\) \[bohr\]\s*$",
                       subMatchers=[SM(r"\s*(?P<x_abinit_atom_xcart>(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){3})\s*$",
                                       repeats=True)]
                       ),
                    SM(startReStr=r"\s*Reduced coordinates \(xred\)\s*$",
                       subMatchers=[SM(r"\s*(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){3}\s*$",
                                       repeats=True)]
                       ),
                    SM(startReStr=r"\s*Cartesian forces \(fcart\) \[Ha/bohr\]; max,rms=(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){2}\s*"
                                  r"\(free atoms\)\s*$",
                       subMatchers=[SM(r"\s*(?P<x_abinit_atom_force>(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){3})\s*$",
                                       repeats=True)]
                       ),
                    SM(startReStr=r"\s*Reduced forces \(fred\)\s*",
                       subMatchers=[SM(r"\s*(\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?){3}\s*$",
                                       repeats=True)]
                       ),
                    SM(r"\s*Total energy \(etotal\) \[Ha\]=\s*(?P<energy_total__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                    SM(r"\s*Difference of energy with previous step \(new-old\):\s*$",
                       coverageIgnore=True),
                    SM(r"\s*Absolute \(Ha\)=[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$"),
                    SM(r"\s*Relative\s*=[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$"),
                    SM(r"fconv : WARNING -\s*$",
                       coverageIgnore=True),
                    SM(r"\s*(At Broyd/MD step\s*\d+\s*, gradients|ntime=\s*\d+\s*was|fconv : at Broyd/MD step\s*\d+\s*, gradients have) "
                       r"(?P<x_abinit_geometry_optimization_converged>(are converged|not enough|not converged)) "
                       r"(:|Broyd/MD steps to converge gradients:|yet.)\s*$"),
                    SM(r"\s*max grad \(force/stress\) =\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*< tolmxf=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*ha/bohr \(free atoms\)\s*")
                    ]
       )


SCFCycleMatcher = \
    SM(name='SCFCycle',
       startReStr=r"(\s*iter\s*(Etot\(hartree\)|2DEtotal\(Ha\))\s*deltaE\((h|Ha)\)(\s*\w+)*|"
                  r"--- Iteration: \(\s*\d+/\d+\) Internal Cycle: \(\d+/\d+\))\s*$",
       repeats=True,
       sections=['section_single_configuration_calculation'],
       subMatchers=[SM(r"-{80}\s*$",
                       coverageIgnore=True),
                    SM(r"---SELF-CONSISTENT-FIELD CONVERGENCE-{44}",
                       coverageIgnore=True),
                    SM(r"(\s*|-)ETOT\s*[0-9]+\s*(?P<energy_total_scf_iteration__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*"
                       r"(?P<energy_change_scf_iteration__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)(\s*[-+0-9.eEdD]*)*",
                       sections=["section_scf_iteration"],
                       repeats=True),
                    SM(r"\s*At SCF step\s*(?P<number_of_scf_iterations>[0-9]+)\s*"
                       r"(, etot|, forces|vres2\s*=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*<\s*tolvrs=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*=>)\s*"
                       r"(?P<x_abinit_single_configuration_calculation_converged>(is converged|are converged|converged))"
                       r"\s*(:|.)\s*$"),
                    SM(r"\s*for the second time, (max diff in force|diff in etot)=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*<\s*tol(dfe|dff)="
                       r"\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$"),
                    SM(startReStr=r"\s*Cartesian components of stress tensor \(hartree/bohr\^3\)\s*$",
                       coverageIgnore=True,
                       sections=["x_abinit_section_stress_tensor"],
                       subMatchers=[SM(r"\s*sigma\(1 1\)=\s*(?P<x_abinit_stress_tensor_xx>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"
                                       r"\s*sigma\(3 2\)=\s*(?P<x_abinit_stress_tensor_zy>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                                    SM(r"\s*sigma\(2 2\)=\s*(?P<x_abinit_stress_tensor_yy>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"
                                       r"\s*sigma\(3 1\)=\s*(?P<x_abinit_stress_tensor_zx>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$"),
                                    SM(r"\s*sigma\(3 3\)=\s*(?P<x_abinit_stress_tensor_zz>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"
                                       r"\s*sigma\(2 1\)=\s*(?P<x_abinit_stress_tensor_yx>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$")
                                    ]
                       ),
                    SM(startReStr=r"\s*Integrated electronic density in atomic spheres:\s*$",
                       required=False,
                       coverageIgnore=True,
                       subMatchers=[SM(r"\s*-{48}\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Atom\s*Sphere_radius\s*Integrated_density\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*\d+\s*[0-9.]+\s*[0-9.]+\s*$",
                                       coverageIgnore=True, repeats=True)
                                    ]
                       ),
                    SM(startReStr=r"\s*Integrated electronic and magnetization densities in atomic spheres:\s*$",
                       required=False,
                       coverageIgnore=True,
                       subMatchers=[SM(r"\s*-{69}\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Note: Diff\(up-dn\) is a rough approximation of local magnetic moment\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Atom\s*Radius\s*up_density\s*dn_density\s*Total\(up\+dn\)\s*Diff\(up-dn\)\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*\d+(\s*[0-9.]+){5}\s*$",
                                       coverageIgnore=True, repeats=True),
                                    SM(r"\s*-{69}\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Sum:(\s*[0-9.]+){4}\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Total magnetization \(from the atomic spheres\):\s*[0-9.]+\s*$",
                                       coverageIgnore=True),
                                    SM(r"\s*Total magnetization \(exact up - dn\):\s*[0-9.]+\s*$",
                                       coverageIgnore=True)
                                    ]
                       ),
                    SCFOutput,
                    SM(r"={80}\s*$",
                       coverageIgnore=True, required=False),
                    SCFResultsMatcher
                    ]
       )


pseudopotentialMatcher = \
    SM(name="pseudopotential",
       startReStr=r"-\s*pspini: atom type\s*[0-9]+\s*psp file is \S*\s*$",
       endReStr=r"\s*pspatm: atomic psp has been read  and splines computed\s*$",
       forwardMatch=True,
       repeats=True,
       coverageIgnore=True,
       subMatchers=[SM(r"-\s*pspini: atom type\s*\d+\s*psp file is\s*\S*\s*$"),
                    SM(r"-\s*pspatm: opening atomic psp file\s*\S*",
                       coverageIgnore=True),
                    SM(r"-(\s*[0-9.]+){3}\s*znucl, zion, pspdat\s*$"),
                    SM(r"(\s*\d+){5}\s*[0-9.]+\s*pspcod,pspxc,lmax,lloc,mmax,r2well\s*$"),
                    SM(startReStr=r"\s*\d+\s*[0-9.]+\s*[0-9.]+\s*\d+\s*[0-9.]+\s*l,e99.0,e99.9,nproj,rcpsp\s*$",
                       repeats=True,
                       subMatchers=[SM(r"\s*([0-9.]+\s*){4}rms, ekb1, ekb2, epsatm\s*$")]),
                    SM(r"\s*([0-9.]+\s*){3}rchrg,fchrg,qchrg\s*$"),
                    SM(startReStr=r"\s*rloc=\s*[0-9.]+\s*$",
                       subMatchers=[SM(r"\s*cc1=\s*[-0-9.]+; cc2=\s*[-0-9.]+; cc3=\s*[-0-9.]+; cc4=\s*[-0-9.]+\s*$"),
                                    SM(r"\s*rrs=\s*[-0-9.]+; h1s=\s*[-0-9.]+; h2s=\s*[-0-9.]+\s*$"),
                                    SM(r"\s*rrp=\s*[-0-9.]+; h1p=\s*[-0-9.]+\s*$"),
                                    SM(r"-  Local part computed in reciprocal space.\s*$")
                                    ]
                       ),
                    SM(r"\s*pspatm : COMMENT -\s*$",
                       coverageIgnore=True),
                    SM(r"\s*the projectors are not normalized,\s*",
                       coverageIgnore=True),
                    SM(r"\s*so that the KB energies are not consistent with\s*$",
                       coverageIgnore=True),
                    SM(r"\s*definition in PRB44, 8503 \(1991\).\s*$",
                       coverageIgnore=True),
                    SM(r"\s*However, this does not influence the results obtained hereafter.\s*$",
                       coverageIgnore=True),
                    SM(r"\s*pspatm: epsatm=\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$"),
                    SM(r"\s*--- l  ekb\(1:nproj\) -->\s*$",
                       coverageIgnore=True),
                    SM(r"\s*\d+\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$",
                       repeats=True)
                    ]
       )

datasetHeaderMatcher = \
    SM(name='DatasetHeader',
       startReStr=r"={2}\s*DATASET\s*[0-9]+\s*={66}\s*$",
       endReStr=r"={80}\s*$",
       forwardMatch=True,
       repeats=False,
       sections=['x_abinit_section_dataset_header'],
       subMatchers=[SM(r"={2}\s*DATASET\s*(?P<x_abinit_dataset_number>[0-9]+)\s*={66}\s*$"),
                    SM(r"-\s*nproc\s*=\s*[0-9]+\s*$"),
                    SM(name="XC",
                       startReStr=r"\s*Exchange-correlation functional for the present dataset will be:",
                       required=False,
                       coverageIgnore=True,
                       subMatchers=[SM(r"(\s*\S*)+\s*-\s*ixc=(?P<x_abinit_var_ixc>[-0-9]+)\s*$"),
                                    SM(r"\s*Citation for XC functional:",
                                       coverageIgnore=True),
                                    SM(r"\s*((\S*.\s*)+\S*,)+\s*\S*\s*[0-9]+,\s*[0-9]+\s*\([0-9]+\)\s*$",
                                       coverageIgnore=True)
                                    ]
                       ),
                    SM(r"\s*Real\(R\)\+Recip\(G\) space primitive vectors, cartesian coordinates \(Bohr,Bohr\^-1\):",
                       coverageIgnore=True),
                    SM(r"\s*R\(1\)=(?P<x_abinit_vprim_1>(\s*[0-9.-]+){3})\s*G\(1\)=(\s*[0-9.-]+){3}\s*$"),
                    SM(r"\s*R\(2\)=(?P<x_abinit_vprim_2>(\s*[0-9.-]+){3})\s*G\(2\)=(\s*[0-9.-]+){3}\s*$"),
                    SM(r"\s*R\(3\)=(?P<x_abinit_vprim_3>(\s*[0-9.-]+){3})\s*G\(3\)=(\s*[0-9.-]+){3}\s*$"),
                    SM(r"\s*Unit cell volume ucvol=\s*[-+0-9.eEdD]*\s*bohr\^3\s*$"),
                    SM(r"\s*Angles \(23,13,12\)=(\s*[-+0-9.eEdD]*){3}\s*degrees\s*$"),
                    SM(r"\s*getcut: wavevector=(\s*[0-9.]*){3}\s*ngfft=(\s*[0-9]*){3}\s*$"),
                    SM(r"\s*ecut\(hartree\)=\s*[0-9.]*\s*=> boxcut\(ratio\)=\s*[0-9.]*\s*$"),
                    SM(r"--- Pseudopotential description ------------------------------------------------",
                       coverageIgnore=True),
                    pseudopotentialMatcher,
                    SM(r"\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*ecore\*ucvol\(ha\*bohr\*\*3\)\s*$",
                       coverageIgnore=True),
                    SM(r"-{80}",
                       coverageIgnore=True),
                    SM(r"P newkpt: treating\s*[0-9]+\s*bands with npw=\s*[0-9]+\s*for ikpt=\s*[0-9]+\s*by node\s*[0-9]+",
                       coverageIgnore=True,
                       repeats=True,
                       required=False),
                    SM(r"_setup2: Arith. and geom. avg. npw \(full set\) are(\s*[0-9.]+\s*){2}",
                       coverageIgnore=True)
                    ]
       )

datasetMatcher = \
    SM(name='Dataset',
       startReStr=r"={2}\s*DATASET\s*[0-9]+\s*={66}\s*$",
       forwardMatch=True,
       repeats=True,
       sections=['x_abinit_section_dataset'],
       subMatchers=[datasetHeaderMatcher,
                    SM("===\s*\[ionmov=\s*\d+\]\s*\S*\s*method\s*\(forces(,Tot energy){0,1}\)\s*$",
                       required=False, coverageIgnore=True),
                    SM(r"={80}\s*$",
                       coverageIgnore=True, required=False, weak=True),
                    SCFCycleMatcher,
                    SM(r"==( END DATASET\(S\) |={16})={62}\s*$",
                       coverageIgnore=True),
                    SM(r"={80}\s*$",
                       coverageIgnore=True, required=False, weak=True)
                    ]
       )


outputVarsMatcher = \
    SM(name='OutputVars',
       startReStr=r"\s*-outvars: echo values of variables after computation  --------",
       endReStr=r"={80}",
       coverageIgnore=True,
       required=True,
       subMatchers=[SM(startReStr=r"[-P]?\s+[_a-zA-Z0-9]+\s+(\S*\s*)+\s*(Hartree|Bohr)?\s*$",
                       repeats=True,
                       coverageIgnore=True,
                       subMatchers=[SM(r"\s{20,}(\S*\s*)+\s*$",
                                       repeats=True, coverageIgnore=True),
                                    SM(r"\s{20,}outvar(_i_n|s)\s*: Printing only first\s*[0-9]*\s*[-a-zA-Z]*.\s*$",
                                       required=False, coverageIgnore=True)
                                    ]
                       )
                    ]
       )

footerMatcher = \
    SM(name='Footer',
       startReStr="\s*Suggested references for the acknowledgment of ABINIT usage.\s*",
       required=True,
       coverageIgnore=True,
       subMatchers=[SM(r"\s*The users of ABINIT have little formal obligations with respect to the ABINIT group",
                       coverageIgnore=True),
                    SM(r"\s*\(those specified in the GNU General Public License, "
                       r"http://www.gnu.org/copyleft/gpl.txt\).",
                       coverageIgnore=True),
                    SM(r"\s*However, it is common practice in the scientific literature,",
                       coverageIgnore=True),
                    SM(r"\s*to acknowledge the efforts of people that have made the research possible.",
                       coverageIgnore=True),
                    SM(r"\s*In this spirit, please find below suggested citations of work written by ABINIT "
                       r"developers,",
                       coverageIgnore=True),
                    SM(r"\s*corresponding to implementations inside of ABINIT that you have used in the present run.",
                       coverageIgnore=True),
                    SM(r"\s*Note also that it will be of great value to readers of publications presenting these "
                       r"results,",
                       coverageIgnore=True),
                    SM(r"\s*to read papers enabling them to understand the theoretical formalism and details",
                       coverageIgnore=True),
                    SM(r"\s*of the ABINIT implementation.",
                       coverageIgnore=True),
                    SM(r"\s*For information on why they are suggested, see also "
                       r"http://www.abinit.org/about/\?text=acknowledgments.",
                       coverageIgnore=True),
                    SM(r"-?(\s*\[[0-9]+\])?(\s*\S*)*",
                       coverageIgnore=True, weak=True, repeats=True),
                    SM(r"- Proc\.\s*[0-9]+\s*individual time \(sec\): cpu=\s*[0-9.]+\s*wall=\s*[0-9.]+\s*",
                       coverageIgnore=True),
                    SM(r"={80}",
                       coverageIgnore=True),
                    SM(r"\s*(?P<x_abinit_completed>Calculation completed)."),
                    SM(r".Delivered\s*[0-9]+\s*WARNINGs and\s*[0-9]+\s*COMMENTs to log file.",
                       coverageIgnore=True),
                    SM(r"\+Overall time at end \(sec\) : cpu=\s*[0-9.]+\s*wall=\s*[0-9.]+",
                       coverageIgnore=True)
                    ]
       )


mainFileDescription = \
    SM(name='root',
       startReStr="\s*$",
       required=True,
       subMatchers=[SM(name='NewRun',
                       startReStr="\s*$",
                       endReStr=r"\s*Overall time at end \(sec\) : cpu=\s*\S*\s*wall=\s*\S*",
                       required=True,
                       fixedStartValues={'program_name': 'ABINIT', 'program_basis_set_type': 'plane waves'},
                       sections=['section_run'],
                       subMatchers=[headerMatcher,
                                    memestimationMatcher,
                                    inputVarsMatcher,
                                    SM(r"={80}", coverageIgnore=True),
                                    datasetMatcher,
                                    outputVarsMatcher,
                                    timerMatcher,
                                    footerMatcher
                                    ]
                       )
                    ]
       )

class AbinitParser():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
       from unittest.mock import patch
       logging.info('abinit parser started')
       logging.getLogger('nomadcore').setLevel(logging.WARNING)
       backend = self.backend_factory("abinit.nomadmetainfo.json")
       with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
           mainFunction(
               mainFileDescription,
               None,
               parserInfo,
               cachingLevelForMetaName = {'x_abinit_section_var': CachingLevel.Cache
                                         },
               superContext=ABINITContext(),
               superBackend=backend)

       return backend



from pathlib import Path
import re


def get_min_ver(dependency: str) -> str:
    """Retrieve version of `dependency` from setup.py, raise if not found."""
    setup_py = Path(__file__).parent.joinpath("../../../setup.py")
    with open(setup_py, "r") as setup_file:
        for line in setup_file.readlines():
            min_ver = re.findall(fr'"{dependency}~=([0-9]+(\.[0-9]+){{,2}})"', line)
            if min_ver:
                return min_ver[0][0]
        else:
            raise RuntimeError(f"Cannot find {dependency} dependency in setup.py")



# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.caching_backend import CachingLevel
from nomadcore.baseclasses import MainHierarchicalParser, CacheService
import re
import logging
import numpy as np
LOGGER = logging.getLogger("nomad")


class NWChemMainParser(MainHierarchicalParser):
    """The main parser class that is called for all run types. Parses the NWChem
    output file.
    """
    def __init__(self, parser_context):
        """
        """
        super(NWChemMainParser, self).__init__(parser_context)

        # Cache for storing current method settings
        self.method_cache = CacheService(self.parser_context)
        self.method_cache.add("single_configuration_to_calculation_method_ref", single=False, update=False)
        self.method_cache.add("program_basis_set_type", single=True, update=False)

        # Cache for storing current sampling method settings
        self.sampling_method_cache = CacheService(self.parser_context)
        self.sampling_method_cache.add("sampling_method", single=False, update=True)
        self.sampling_method_cache.add("ensemble_type", single=False, update=True)

        # Cache for storing frame sequence information
        self.frame_sequence_cache = CacheService(self.parser_context)
        self.frame_sequence_cache.add("number_of_frames_in_sequence", 0, single=False, update=True)
        self.frame_sequence_cache.add("frame_sequence_local_frames_ref", [], single=False, update=True)
        self.frame_sequence_cache.add("frame_sequence_potential_energy", [], single=False, update=True)
        self.frame_sequence_cache.add("frame_sequence_kinetic_energy", [], single=False, update=True)
        self.frame_sequence_cache.add("frame_sequence_temperature", [], single=False, update=True)
        self.frame_sequence_cache.add("frame_sequence_time", [], single=False, update=True)
        self.frame_sequence_cache.add("frame_sequence_to_sampling_ref", single=False, update=True)

        # Cache for storing system information
        self.system_cache = CacheService(self.parser_context)
        self.system_cache.add("initial_positions", single=False, update=False)
        self.system_cache.add("atom_positions", single=False, update=True)
        self.system_cache.add("atom_labels", single=False, update=False)
        self.system_cache.add("configuration_periodic_dimensions", single=False, update=False)

        # Cache for storing scc information
        self.scc_cache = CacheService(self.parser_context)
        self.scc_cache.add("atom_forces", single=False, update=True)
        self.scc_cache.add("number_of_scf_iterations", 0, single=False, update=True)
        self.scc_cache.add("single_configuration_calculation_to_system_ref", single=False, update=True)

        #=======================================================================
        # Cache levels
        self.caching_levels.update({
            'x_nwchem_section_geo_opt_module': CachingLevel.Cache,
            'x_nwchem_section_geo_opt_step': CachingLevel.Cache,
            'x_nwchem_section_xc_functional': CachingLevel.Cache,
            'x_nwchem_section_qmd_module': CachingLevel.ForwardAndCache,
            'x_nwchem_section_qmd_step': CachingLevel.ForwardAndCache,
            'x_nwchem_section_xc_part': CachingLevel.ForwardAndCache,
        })

        #=======================================================================
        # Main Structure
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=['section_run'],
            subMatchers=[
                self.input(),
                self.header(),
                self.system(),

                # This repeating submatcher supports multiple different tasks
                # within one run
                SM("(?:\s+NWChem DFT Module)|(?:\s+NWChem Geometry Optimization)|(?:\s+NWChem QMD Module)|(?:\s+\*               NWPW PSPW Calculation              \*)",
                    repeats=True,
                    forwardMatch=True,
                    subFlags=SM.SubFlags.Unordered,
                    subMatchers=[
                        self.energy_force_gaussian_task(),
                        self.energy_force_pw_task(),
                        self.gaussian_geo_opt_module(),
                        self.dft_gaussian_md_task(),
                    ]
                ),
            ]
        )

    def input(self):
        """Returns the simplematcher that parses the NWChem input
        """
        return SM( re.escape("============================== echo of input deck =============================="),
            endReStr=re.escape("================================================================================"),
        )

    def system(self):
        return SM( "                                NWChem Input Module",
            subMatchers=[
                self.geometry(),
                SM( r"  No\.       Tag          Charge          X              Y              Z"),
                SM( re.escape(r" ---- ---------------- ---------- -------------- -------------- --------------"),
                    adHoc=self.adHoc_atoms,
                ),
            ]
        )

    def energy_force_gaussian_task(self):
        return SM( "                                 NWChem DFT Module",
            startReAction=self.set_gaussian_basis,
            forwardMatch=True,
            subMatchers=[
                self.dft_calculation_full(),
            ],

        )

    def energy_force_pw_task(self):
        return SM( "          \*               NWPW PSPW Calculation              \*",
            sections=["section_single_configuration_calculation", "section_system", "section_method"],
            startReAction=self.set_plane_wave_basis,
            fixedStartValues={
                "electronic_structure_method": "DFT",
            },
            subMatchers=[
                SM(" number of processors used:\s+{}".format(self.regexs.int)),
                SM(" processor grid           :\s+{0} x\s+{0}".format(self.regexs.int)),
                SM(" parallel mapping         :{}".format(self.regexs.eol)),
                SM(" parallel mapping         :{}".format(self.regexs.eol)),
                SM(" number of threads        :{}".format(self.regexs.eol)),
                SM(" parallel io              :{}".format(self.regexs.eol)),
                SM(" options:",
                    subMatchers=[
                        SM("      boundary conditions  =\s+({})".format(self.regexs.word),
                            startReAction=self.transform_periodicity
                        ),
                        SM("      electron spin        =\s+(?P<x_nwchem_electron_spin_restriction>{})".format(self.regexs.word),
                        ),
                        SM("      exchange-correlation =\s+({})".format(self.regexs.eol),
                            startReAction=self.transform_xc
                        ),
                    ]
                ),
                SM(" total charge:\s+({})".format(self.regexs.float),
                    startReAction=self.transform_total_charge,
                ),

                SM(" supercell:",
                    subMatchers=[
                        SM("      cell_name:\s+{}".format(self.regexs.eol)),
                        SM("      lattice:\s+a1=<",
                            forwardMatch=True,
                            adHoc=self.adHoc_lattice("simulation_cell", "angstrom")
                        ),
                        SM("      lattice:\s+a1=<",
                            forwardMatch=True,
                            adHoc=self.adHoc_lattice("x_nwchem_reciprocal_simulation_cell", "angstrom^-1")
                        ),
                        SM("      lattice:\s+a1=<",
                            adHoc=self.adHoc_lattice_parameters
                        ),
                    ]
                ),
            ],
        )

    def gaussian_geo_opt_module(self):
        return SM( "                           NWChem Geometry Optimization",
            startReAction=self.set_gaussian_basis,
            sections=["section_frame_sequence", "section_sampling_method", "x_nwchem_section_geo_opt_module"],
            subFlags=SM.SubFlags.Sequenced,
            subMatchers=[
                SM( r" maximum gradient threshold         \(gmax\) =\s+(?P<geometry_optimization_threshold_force__forceAu>{})".format(self.regexs.float)),
                SM( r" rms gradient threshold             \(grms\) =\s+{}".format(self.regexs.float)),
                SM( r" maximum cartesian step threshold   \(xmax\) =\s+(?P<geometry_optimization_geometry_change__bohr>{})".format(self.regexs.float)),
                SM( r" rms cartesian step threshold       \(xrms\) =\s+{}".format(self.regexs.float)),
                SM( r" fixed trust radius                \(trust\) =\s+{}".format(self.regexs.float)),
                SM( r" maximum step size to saddle      \(sadstp\) =\s+{}".format(self.regexs.float)),
                SM( r" energy precision                  \(eprec\) =\s+(?P<geometry_optimization_energy_change__hartree>{})".format(self.regexs.float)),
                SM( r" maximum number of steps          \(nptopt\) =\s+{}".format(self.regexs.int)),
                SM( r" initial hessian option           \(inhess\) =\s+{}".format(self.regexs.int)),
                SM( r" line search option               \(linopt\) =\s+{}".format(self.regexs.int)),
                SM( r" hessian update option            \(modupd\) =\s+{}".format(self.regexs.int)),
                SM( r" saddle point option              \(modsad\) =\s+{}".format(self.regexs.int)),
                SM( r" initial eigen-mode to follow     \(moddir\) =\s+{}".format(self.regexs.int)),
                SM( r" initial variable to follow       \(vardir\) =\s+{}".format(self.regexs.int)),
                SM( r" follow first negative mode     \(firstneg\) =\s+{}".format(self.regexs.word)),
                SM( r" apply conjugacy                    \(opcg\) =\s+{}".format(self.regexs.word)),
                SM( r" source of zmatrix                         =\s+{}".format(self.regexs.word)),
                SM( r"          Energy Minimization"),
                SM( r" Names of Z-matrix variables"),
                SM( "          Step\s+\d+$",
                    sections=["x_nwchem_section_geo_opt_step"],
                    subMatchers=[
                        self.geometry(),
                        self.dft_calculation_full(on_close_scc=self.add_frame_reference),
                        SM( "[.@] Step       Energy      Delta E   Gmax     Grms     Xrms     Xmax   Walltime"),
                        SM( "[.@] ---- ---------------- -------- -------- -------- -------- -------- --------"),
                        SM( "@\s+{0}\s+(?P<x_nwchem_geo_opt_step_energy__hartree>{1})\s+{1}\s+{1}\s+{1}\s+{1}\s+{1}\s+{1}".format(self.regexs.int, self.regexs.float)),
                        self.dft_calculation_no_method(),
                    ]
                ),
                SM("          Step\s+\d+$",
                    repeats=True,
                    forwardMatch=True,
                    sections=["x_nwchem_section_geo_opt_step"],
                    subMatchers=[
                        SM("          Step\s+\d+$"),
                        self.geometry(),
                        self.dft_calculation_no_method(on_close_scc=self.add_frame_reference),
                        SM( "[.@] Step       Energy      Delta E   Gmax     Grms     Xrms     Xmax   Walltime"),
                        SM( "[.@] ---- ---------------- -------- -------- -------- -------- -------- --------"),
                        SM( "@\s+{0}\s+(?P<x_nwchem_geo_opt_step_energy__hartree>{1})\s+{1}\s+{1}\s+{1}\s+{1}\s+{1}\s+{1}".format(self.regexs.int, self.regexs.float)),
                        self.dft_calculation_no_method(),
                    ],
                ),
                SM( "\s+Optimization converged",
                    fixedStartValues={
                        "geometry_optimization_converged": True
                    }
                )
            ]
        )

    def dft_gaussian_md_task(self):
        return SM( "                                 NWChem QMD Module",
            startReAction=self.set_gaussian_basis,
            sections=["section_frame_sequence", "section_sampling_method", "x_nwchem_section_qmd_module"],
            subMatchers=[
                SM("                                QMD Run Parameters",
                    subMatchers=[
                        SM("    No. of nuclear steps:\s+(?P<x_nwchem_qmd_number_of_nuclear_steps>{})".format(self.regexs.int)),
                        SM("       Nuclear time step:\s+(?P<x_nwchem_qmd_nuclear_time_step>{})".format(self.regexs.float)),
                        SM("        Target temp\. \(K\):\s+(?P<x_nwchem_qmd_target_temperature>{})".format(self.regexs.float)),
                        SM("              Thermostat:\s+(?P<x_nwchem_qmd_thermostat>{})".format(self.regexs.eol)),
                        SM("                     Tau:\s+(?P<x_nwchem_qmd_tau>{})".format(self.regexs.float)),
                        SM("             Random seed:\s+(?P<x_nwchem_qmd_random_seed>{})".format(self.regexs.int)),
                        SM("      Nuclear integrator:\s+(?P<x_nwchem_qmd_nuclear_integrator>{})".format(self.regexs.eol)),
                        SM("       Current temp. \(K\):\s+(?P<x_nwchem_qmd_initial_temperature__K>{})".format(self.regexs.float)),
                    ]
                ),
                self.dft_calculation_full(),
                SM("                                 NWChem DFT Module",
                    repeats=True,
                    sections=[
                        "x_nwchem_section_qmd_step",
                        "section_single_configuration_calculation",
                        "section_system"
                    ],
                    onClose={
                        "section_single_configuration_calculation": self.add_frame_reference,
                        "section_system": self.push_no_periodicity
                    },
                    subMatchers=[
                        SM("  Caching 1-el integrals"),
                        SM("   Time after variat\. SCF:\s+{}".format(self.regexs.float)),
                        SM("   Time prior to 1st pass:\s+{}".format(self.regexs.float)),
                        SM("         Total DFT energy =\s+(?P<energy_total__hartree>{})".format(self.regexs.float)),
                        SM("      One electron energy =\s+(?P<x_nwchem_energy_one_electron__hartree>{})".format(self.regexs.float)),
                        SM("           Coulomb energy =\s+(?P<x_nwchem_energy_coulomb__hartree>{})".format(self.regexs.float)),
                        SM("    Exchange-Corr\. energy =\s+(?P<energy_XC__hartree>{})".format(self.regexs.float)),
                        SM(" Nuclear repulsion energy =\s+(?P<x_nwchem_energy_nuclear_repulsion__hartree>{})".format(self.regexs.float)),
                        SM(" Numeric\. integr\. density =\s+{}".format(self.regexs.float)),
                        SM("     Total iterative time =\s+(?P<time_calculation__s>{})".format(self.regexs.float)),
                        SM("                         DFT ENERGY GRADIENTS",
                            subMatchers=[
                                SM("    atom               coordinates                        gradient"),
                                SM("                 x          y          z           x          y          z",
                                    adHoc=self.adHoc_forces(save_positions=True)
                                ),
                            ]
                        ),
                        SM("            QMD Run Information",
                            subMatchers=[
                                SM("  Time elapsed \(fs\) :\s+(?P<x_nwchem_qmd_step_time__fs>{})".format(self.regexs.float)),
                                SM("  Kin. energy \(a\.u\.\):\s+{}\s+(?P<x_nwchem_qmd_step_kinetic_energy__hartree>{})".format(self.regexs.int, self.regexs.float)),
                                SM("  Pot. energy \(a\.u\.\):\s+{}\s+(?P<x_nwchem_qmd_step_potential_energy__hartree>{})".format(self.regexs.int, self.regexs.float)),
                                SM("  Tot. energy \(a\.u\.\):\s+{}\s+(?P<x_nwchem_qmd_step_total_energy__hartree>{})".format(self.regexs.int, self.regexs.float)),
                                SM("  Target temp\. \(K\)  :\s+{}\s+(?P<x_nwchem_qmd_step_target_temperature__K>{})".format(self.regexs.int, self.regexs.float)),
                                SM("  Current temp\. \(K\) :\s+{}\s+(?P<x_nwchem_qmd_step_temperature__K>{})".format(self.regexs.int, self.regexs.float)),
                                SM("  Dipole \(a\.u\.\)     :\s+{0}\s+({1}\s+{1}\s+{1})".format(self.regexs.int, self.regexs.float), startReAction=self.transform_dipole)
                            ]
                        )
                    ]
                )
            ]
        )

    def geometry(self):
        return SM( r"                         Geometry \"geometry\" -> \"geometry\"",
            sections=["x_nwchem_section_geometry"],
            subMatchers=[
                SM(r"                         ---------------------------------"),
                SM(r" Output coordinates in angstroms \(scale by\s+{}to convert to a\.u\.\)"),
                SM(r"  No\.       Tag          Charge          X              Y              Z"),
                SM(r" ---- ---------------- ---------- -------------- -------------- --------------",
                    adHoc=self.adHoc_atoms),
            ]
        )

    def header(self):
        """Returns the simplematcher that parser the NWChem header
        """
        return SM( "\s+Northwest Computational Chemistry Package \(NWChem\) (?P<program_version>(\d+\.\d+(?:\.\d+)?))",
            sections=["x_nwchem_section_start_information"],
            subMatchers=[
                SM( r"\s+hostname\s+= (?P<x_nwchem_run_host_name>{})".format(self.regexs.eol)),
                SM( r"\s+program\s+= (?P<x_nwchem_program_name>{})".format(self.regexs.eol)),
                SM( r"\s+date\s+= (?P<x_nwchem_start_datetime>{})".format(self.regexs.eol)),
                SM( r"\s+compiled\s+= (?P<x_nwchem_compilation_datetime>{})".format(self.regexs.eol)),
                SM( r"\s+compiled\s+= (?P<x_nwchem_compilation_datetime>{})".format(self.regexs.eol)),
                SM( r"\s+source\s+= (?P<x_nwchem_source>{})".format(self.regexs.eol)),
                SM( r"\s+nwchem branch\s+= (?P<x_nwchem_branch>{})".format(self.regexs.eol)),
                SM( r"\s+nwchem revision\s+= (?P<x_nwchem_revision>{})".format(self.regexs.eol)),
                SM( r"\s+ga revision\s+= (?P<x_nwchem_ga_revision>{})".format(self.regexs.eol)),
                SM( r"\s+input\s+= (?P<x_nwchem_input_filename>{})".format(self.regexs.eol)),
                SM( r"\s+prefix\s+= (?P<x_nwchem_input_prefix>{})".format(self.regexs.eol)),
                SM( r"\s+data base\s+= (?P<x_nwchem_db_filename>{})".format(self.regexs.eol)),
                SM( r"\s+status\s+= (?P<x_nwchem_status>{})".format(self.regexs.eol)),
                SM( r"\s+nproc\s+= (?P<x_nwchem_nproc>{})".format(self.regexs.eol)),
                SM( r"\s+time left\s+= (?P<x_nwchem_time_left>{})".format(self.regexs.eol)),
            ]
        )

    def dft_calculation_no_method(self, on_close_scc=None):
        """SimpleMatcher that ignores DFT method settings but parses actual
        calculation. Used in MD and geo opt, when the settings have not changed
        to avoid storing redundant data.
        """
        return SM( "                                 NWChem DFT Module",
            sections=[
                "section_single_configuration_calculation",
                "section_system",
            ],
            onClose={
                "section_single_configuration_calculation": on_close_scc,
                "section_system": self.push_no_periodicity
            },
            subMatchers=self.dft_calculation_data(),
        )

    def dft_calculation_full(self, on_close_scc=None):
        subMatchers = [
            SM( r"          No. of atoms     :\s+(?P<number_of_atoms>{})".format(self.regexs.int)),
            SM( r"          Charge           :\s+(?P<total_charge>{})".format(self.regexs.int)),
            SM( r"          Spin multiplicity:\s+(?P<spin_target_multiplicity>{})".format(self.regexs.int)),
            SM( r"          Maximum number of iterations:\s+(?P<scf_max_iteration>{})".format(self.regexs.int)),
            SM( r"          Convergence on energy requested:\s+(?P<scf_threshold_energy_change__hartree>{})".format(self.regexs.float)),
            SM( r"          Convergence on density requested:\s+{}".format(self.regexs.float)),
            SM( r"          Convergence on gradient requested:\s+{}".format(self.regexs.float)),
            SM( r"              XC Information",
                subFlags=SM.SubFlags.Unordered,
                subMatchers=[
                    SM("\s+(?P<x_nwchem_xc_functional_shortcut>B3LYP Method XC Potential)"),
                    SM("\s+(?P<x_nwchem_xc_functional_shortcut>PBE0 Method XC Functional)"),
                    SM("\s+(?P<x_nwchem_xc_functional_shortcut>Becke half-and-half Method XC Potential)"),
                    SM("\s+(?P<x_nwchem_xc_functional_shortcut>HCTH120  Method XC Functional)"),
                    SM("\s+(?P<x_nwchem_xc_functional_shortcut>HCTH147  Method XC Functional)"),
                    SM("\s+(?P<x_nwchem_xc_functional_shortcut>HCTH407 Method XC Functional)"),
                    SM("\s+(?P<x_nwchem_xc_functional_name>PerdewBurkeErnzerhof Exchange Functional)\s+(?P<x_nwchem_xc_functional_weight>{})".format(self.regexs.float), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Becke 1988 Exchange Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Lee-Yang-Parr Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Perdew 1991   Exchange Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Perdew 1991 Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Perdew 1991 LDA Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>PerdewBurkeErnz. Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Perdew 1981 Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Perdew 1986 Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Perdew 1991 Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Hartree-Fock \(Exact\) Exchange)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>Slater Exchange Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>OPTX     Exchange Functional)\s+(?P<x_nwchem_xc_functional_weight>{})\s+(?P<x_nwchem_xc_functional_type>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>TPSS metaGGA Exchange Functional)\s+(?P<x_nwchem_xc_functional_weight>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>TPSS03 metaGGA Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                    SM("\s+(?P<x_nwchem_xc_functional_name>VWN V Correlation Functional)\s+(?P<x_nwchem_xc_functional_weight>{})".format(self.regexs.float, self.regexs.eol), sections=["x_nwchem_section_xc_part"]),
                ],
            ),
        ]
        subMatchers.extend(self.dft_calculation_data())

        return SM( "                                 NWChem DFT Module",
            sections=[
                "section_single_configuration_calculation",
                "section_system",
                "section_method"
            ],
            fixedStartValues={"electronic_structure_method": "DFT"},
            onClose={
                "section_single_configuration_calculation": on_close_scc,
                "section_system": self.push_no_periodicity
            },
            subMatchers=subMatchers,
        )

    def dft_calculation_data(self):
        return [
            SM( r"   convergence    iter        energy       DeltaE   RMS-Dens  Diis-err    time",
                subMatchers=[
                    SM( r" d=\s+{1},ls={0},diis\s+{1}\s+(?P<energy_total_scf_iteration__hartree>{0})\s+(?P<energy_change_scf_iteration__hartree>{0})\s+{0}\s+{0}\s+{0}".format(self.regexs.float, self.regexs.int),
                        sections=["section_scf_iteration"],
                        repeats=True,
                    )
                ]
            ),
            SM( r"         Total DFT energy =\s+(?P<energy_total__hartree>{})".format(self.regexs.float)),
            SM( r"      One electron energy =\s+(?P<x_nwchem_energy_one_electron__hartree>{})".format(self.regexs.float)),
            SM( r"           Coulomb energy =\s+(?P<x_nwchem_energy_coulomb__hartree>{})".format(self.regexs.float)),
            SM( r"          Exchange energy =\s+(?P<energy_X__hartree>{})".format(self.regexs.float)),
            SM( r"       Correlation energy =\s+(?P<energy_C__hartree>{})".format(self.regexs.float)),
            SM( r" Nuclear repulsion energy =\s+(?P<x_nwchem_energy_nuclear_repulsion__hartree>{})".format(self.regexs.float)),
            self.dft_gradient_module(),
        ]

    def dft_gradient_module(self):
        return SM( r"                            NWChem DFT Gradient Module",
            subMatchers=[
                SM( r"                         DFT ENERGY GRADIENTS"),
                SM( r"    atom               coordinates                        gradient"),
                SM( r"                 x          y          z           x          y          z",
                    adHoc=self.adHoc_forces(),
                ),
            ],
        )

    #=======================================================================
    # onClose triggers
    def onClose_section_run(self, backend, gIndex, section):
        backend.addValue("program_name", "NWChem")
        self.method_cache.addValue("program_basis_set_type")
        # backend.addValue("program_basis_set_type", "gaussians+plane_waves")

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        self.scc_cache.addValue("single_configuration_calculation_to_system_ref")
        self.method_cache.addValue("single_configuration_to_calculation_method_ref")

        if self.scc_cache["atom_forces"] is not None:
            self.scc_cache.addArrayValues("atom_forces", unit="hartree / bohr")

        if self.scc_cache["number_of_scf_iterations"]:
            self.scc_cache.addValue("number_of_scf_iterations")

        self.scc_cache.clear()

    def onClose_section_method(self, backend, gIndex, section):

        # XC settings
        class XCFunctional(object):
            def __init__(self, name, weight, locality=None):
                self.name = name
                self.weight = weight
                self.locality = locality

            def __eq__(self, other):
                if isinstance(other, self.__class__):
                    return self.__dict__ == other.__dict__
                else:
                    return False

            def get_key(self):
                return "{}_{}_{}".format(self.name, self.weight, self.locality)

        xc_final_list = []

        # Check if shortcut was defined
        shortcut = section.get_latest_value("x_nwchem_xc_functional_shortcut")
        if shortcut:
            shortcut_map = {
                "B3LYP Method XC Potential": "HYB_GGA_XC_B3LYP",
                "PBE0 Method XC Functional": "HYB_GGA_XC_PBEH",
                "Becke half-and-half Method XC Potential": "HYB_GGA_XC_BHANDH",
                "HCTH120  Method XC Functional": "GGA_XC_HCTH_120",
                "HCTH147  Method XC Functional": "GGA_XC_HCTH_147",
                "HCTH407 Method XC Functional": "GGA_XC_HCTH_407",
            }
            norm_name = shortcut_map.get(shortcut)
            if norm_name:
                xc_final_list.append(XCFunctional(norm_name, 1.0))
        else:
            # Check if any combination with a more generic name is present
            functionals = section["x_nwchem_section_xc_part"]
            if functionals:
                xc_info = {}
                for functional in functionals:
                    name = functional.get_latest_value("x_nwchem_xc_functional_name")
                    weight = functional.get_latest_value("x_nwchem_xc_functional_weight")
                    locality = functional.get_latest_value("x_nwchem_xc_functional_type")
                    if locality is not None:
                        locality = locality.strip()
                    xc = XCFunctional(name, weight, locality)
                    xc_info[xc.get_key()] = xc

                # Check if a predetermined combination of
                # x_nwchem_section_xc_parts are found.
                combinations = {
                    "GGA_X_OPTX": (
                        XCFunctional("OPTX     Exchange Functional", "1.432", "non-local"),
                        XCFunctional("Slater Exchange Functional", "1.052", "local")
                    ),
                    "GGA_C_PBE": (
                        XCFunctional("Perdew 1991 LDA Correlation Functional", "1.0", "local"),
                        XCFunctional("PerdewBurkeErnz. Correlation Functional", "1.0", "non-local")
                    ),
                    "GGA_C_P86": (
                        XCFunctional("Perdew 1981 Correlation Functional", "1.0", "local"),
                        XCFunctional("Perdew 1986 Correlation Functional", "1.0", "non-local")
                    ),
                    "GGA_C_PW91": (
                        XCFunctional("Perdew 1991 Correlation Functional", "1.0", "non-local"),
                        XCFunctional("Perdew 1991 LDA Correlation Functional", "1.0", "local")
                    ),
                }
                for name, parts in combinations.items():
                    combination_found = True
                    for part in parts:
                        if part.get_key() not in xc_info:
                            combination_found = False

                    if combination_found:
                        for part in parts:
                            del xc_info[part.get_key()]
                        xc = XCFunctional(name, 1.0)
                        xc_final_list.append(xc)

                # Gather the pieces that were not part of any bigger
                # combination
                for xc in xc_info.values():
                    component_map = {
                        "PerdewBurkeErnzerhof Exchange Functional": "GGA_X_PBE",
                        "Becke 1988 Exchange Functional": "GGA_X_B88",
                        "Lee-Yang-Parr Correlation Functional": "GGA_C_LYP",
                        "Perdew 1986 Correlation Functional": "GGA_C_P86",
                        "Perdew 1991 Correlation Functional": "GGA_C_PW91",
                        "Perdew 1991   Exchange Functional": "GGA_X_PW91",
                        "Hartree-Fock \(Exact\) Exchange": "HF_X",
                        "TPSS metaGGA Exchange Functional": "MGGA_X_TPSS",
                        "TPSS03 metaGGA Correlation Functional": "MGGA_C_TPSS",
                        "Slater Exchange Functional": "LDA_X",
                        "VWN V Correlation Functional": "LDA_C_VWN",
                    }
                    name = xc.name
                    locality = xc.locality
                    weight = xc.weight
                    norm_name = component_map.get(name)
                    if norm_name:
                        xc = XCFunctional(norm_name, weight)
                        xc_final_list.append(xc)

        # Go throught the list of found functionals and push XC sections and
        # gather the summary string.
        xc_final_list.sort(key=lambda x: x.name)
        xc_summary = ""
        for i_xc, xc in enumerate(xc_final_list):
            if i_xc != 0:
                xc_summary += "+"
            xc_summary += "{}*{}".format(xc.weight, xc.name)

            # Push the XC sections
            id_xc = backend.openSection("section_XC_functionals")
            backend.addValue("XC_functional_name", xc.name)
            if xc.weight is not None:
                backend.addValue("XC_functional_weight", xc.weight)
            backend.closeSection("section_XC_functionals", id_xc)

        # Push the summary string
        if xc_summary is not "":
            self.backend.addValue("XC_functional", xc_summary)

    def onClose_section_sampling_method(self, backend, gIndex, section):

        thermostat = section.get_latest_value("x_nwchem_qmd_thermostat")
        ensemble = None
        if thermostat == "svr":
            ensemble = "NVT"
            self.backend.addValue("ensemble_type", ensemble)

        self.sampling_method_cache.addValue("sampling_method")
        self.sampling_method_cache.clear()

    def onClose_section_system(self, backend, gIndex, section):
        if self.system_cache["atom_positions"] is not None:
            self.system_cache.addArrayValues("atom_positions", unit="bohr")
        elif self.system_cache["initial_positions"] is not None:
            self.system_cache.addArrayValues("atom_positions", "initial_positions", unit="angstrom")
        self.system_cache.addArrayValues("atom_labels")
        self.system_cache.addArrayValues("configuration_periodic_dimensions")

    def onClose_section_frame_sequence(self, backend, gIndex, section):
        self.frame_sequence_cache.addValue("number_of_frames_in_sequence")
        frame_sequence = self.frame_sequence_cache["frame_sequence_local_frames_ref"]
        if frame_sequence:
            frame_sequence = np.array(frame_sequence)
            self.backend.addArrayValues("frame_sequence_local_frames_ref", frame_sequence)
        self.frame_sequence_cache.addValue("frame_sequence_to_sampling_ref")

        potential_energy = self.frame_sequence_cache["frame_sequence_potential_energy"]
        if potential_energy:
            potential_energy = np.array(potential_energy)
            backend.addArrayValues("frame_sequence_potential_energy", potential_energy)
            backend.addArrayValues("frame_sequence_potential_energy_stats", np.array([potential_energy.mean(), potential_energy.std()]))

        kin_energy = self.frame_sequence_cache["frame_sequence_kinetic_energy"]
        if kin_energy:
            kin_energy = np.array(kin_energy)
            backend.addArrayValues("frame_sequence_kinetic_energy", kin_energy)
            backend.addArrayValues("frame_sequence_kinetic_energy_stats", np.array([kin_energy.mean(), kin_energy.std()]))

        temp = self.frame_sequence_cache["frame_sequence_temperature"]
        if temp:
            temp = np.array(temp)
            backend.addArrayValues("frame_sequence_temperature", temp)
            backend.addArrayValues("frame_sequence_temperature_stats", np.array([temp.mean(), temp.std()]))

        time = self.frame_sequence_cache["frame_sequence_time"]
        if time:
            time = np.array(time)
            backend.addArrayValues("frame_sequence_time", time)

        self.frame_sequence_cache.clear()

    def onClose_section_scf_iteration(self, backend, gIndex, section):
        self.scc_cache["number_of_scf_iterations"] += 1

    def onClose_x_nwchem_section_qmd_step(self, backend, gIndex, section):
        self.frame_sequence_cache["number_of_frames_in_sequence"] += 1

        potential_energy = section.get_latest_value("x_nwchem_qmd_step_potential_energy")
        if potential_energy is not None:
            self.frame_sequence_cache["frame_sequence_potential_energy"].append(potential_energy)

        kin_energy = section.get_latest_value("x_nwchem_qmd_step_kinetic_energy")
        if kin_energy is not None:
            self.frame_sequence_cache["frame_sequence_kinetic_energy"].append(kin_energy)

        temp = section.get_latest_value("x_nwchem_qmd_step_temperature")
        if temp is not None:
            self.frame_sequence_cache["frame_sequence_temperature"].append(temp)

        time = section.get_latest_value("x_nwchem_qmd_step_time")
        if time is not None:
            self.frame_sequence_cache["frame_sequence_time"].append(time)

    def onClose_x_nwchem_section_geo_opt_step(self, backend, gIndex, section):
        self.frame_sequence_cache["number_of_frames_in_sequence"] += 1
        pot_ener = section.get_latest_value("x_nwchem_geo_opt_step_energy")
        if pot_ener is not None:
            self.frame_sequence_cache["frame_sequence_potential_energy"].append(pot_ener)

    #=======================================================================
    # onOpen triggers
    def onOpen_section_method(self, backend, gIndex, section):
        self.method_cache["single_configuration_to_calculation_method_ref"] = gIndex

    def onOpen_section_system(self, backend, gIndex, section):
        self.scc_cache["single_configuration_calculation_to_system_ref"] = gIndex

    def onOpen_section_sampling_method(self, backend, gIndex, section):
        self.frame_sequence_cache["frame_sequence_to_sampling_ref"] = gIndex

    def onOpen_x_nwchem_section_qmd_module(self, backend, gIndex, section):
        self.sampling_method_cache["sampling_method"] = "molecular_dynamics"

    def onOpen_x_nwchem_section_geo_opt_module(self, backend, gIndex, section):
        self.sampling_method_cache["sampling_method"] = "geometry_optimization"

    #=======================================================================
    # adHoc
    def adHoc_forces(self, save_positions=False):
        def wrapper(parser):
            match = True
            forces = []
            positions = []

            while match:
                line = parser.fIn.readline()
                if line == "" or line.isspace():
                    match = False
                    break
                components = line.split()
                position = np.array([float(x) for x in components[-6:-3]])
                force = np.array([float(x) for x in components[-3:]])
                forces.append(force)
                positions.append(position)

            forces = -np.array(forces)
            positions = np.array(positions)

            # If anything found, push the results to the correct section
            if forces.size != 0:
                self.scc_cache["atom_forces"] = forces
            if save_positions:
                if positions.size != 0:
                    self.system_cache["atom_positions"] = positions
        return wrapper

    def adHoc_atoms(self, parser):
        # Define the regex that extracts the information
        regex_string = r"\s+({0})\s+({1})\s+({2})\s+({2})\s+({2})\s+({2})".format(self.regexs.int, self.regexs.word, self.regexs.float)
        regex_compiled = re.compile(regex_string)

        match = True
        coordinates = []
        labels = []

        while match:
            line = parser.fIn.readline()
            result = regex_compiled.match(line)

            if result:
                match = True
                results = result.groups()
                label = results[1]
                labels.append(label)
                coordinate = [float(x) for x in results[3:6]]
                coordinates.append(coordinate)
            else:
                match = False
        coordinates = np.array(coordinates)
        labels = np.array(labels)

        # If anything found, push the results to the correct section
        if len(coordinates) != 0:
            self.system_cache["initial_positions"] = coordinates
            self.system_cache["atom_labels"] = labels

    def adHoc_lattice(self, metaname, units):
        def wrapper(parser):
            a = [float(x) for x in parser.fIn.readline().split()[2:5]]
            b = [float(x) for x in parser.fIn.readline().split()[1:4]]
            c = [float(x) for x in parser.fIn.readline().split()[1:4]]

            cell = np.array([a, b, c])
            parser.backend.addArrayValues(metaname, cell, unit=units)
        return wrapper

    def adHoc_lattice_parameters(self, parser):
        pass

    #=======================================================================
    # SimpleMatcher specific onClose
    def save_geo_opt_sampling_id(self, backend, gIndex, section):
        backend.addValue("frame_sequence_to_sampling_ref", gIndex)

    def add_frame_reference(self, backend, gIndex, section):
        self.frame_sequence_cache["frame_sequence_local_frames_ref"].append(gIndex)

    def push_no_periodicity(self, backend, gIndex, section):
        self.system_cache["configuration_periodic_dimensions"] = np.array([False, False, False])

    #=======================================================================
    # Start match transforms
    def transform_dipole(self, backend, groups):
        dipole = groups[0]
        components = np.array([float(x) for x in dipole.split()])
        backend.addArrayValues("x_nwchem_qmd_step_dipole", components)

    def transform_periodicity(self, backend, groups):
        periodicity = groups[0]
        if periodicity == "periodic":
            self.system_cache["configuration_periodic_dimensions"] = np.array([True, True, True])

    def transform_xc(self, backend, groups):
        xc = groups[0]
        if xc == "LDA (Vosko et al) parameterization":
            pass

    def transform_total_charge(self, backend, groups):
        charge = groups[0]
        self.backend.addValue("total_charge", int(float(charge)))

    def set_gaussian_basis(self, backend, groups):
        self.method_cache["program_basis_set_type"] = "gaussians"

    def set_plane_wave_basis(self, backend, groups):
        self.method_cache["program_basis_set_type"] = "plane waves"

    #=======================================================================
    # Misc
    def debug_end(self):
        def wrapper():
            print("DEBUG END")
        return wrapper

    def debug_close(self):
        def wrapper(backend, gIndex, section):
            print("DEBUG CLOSE")
        return wrapper
